#include <stdio.h>
#include <math.h>
#include <assert.h>

#include "CCamera.h"
#include "fixed.h"
#include "CGlobal.h"

#define DEFAULT_P       512
#define DEFAULT_THETA   (30*256/360)
#define DEFAULT_PHI     (45*256/360)
#define DEFAULT_GAMMA   0
#define DEFAULT_EYEDIST 256

#define SCR_X 320     // the screen resolution, obviously
#define SCR_Y 200
#define SQR(X)          ((X)*(X))

extern CGlobal Global;

CCamera::CCamera( )
  : m_vCenter( 0, 0, 0 ),
    m_p( DEFAULT_P ),
    m_theta( DEFAULT_THETA<<F_SHIFT ),
    m_gamma( DEFAULT_GAMMA<<F_SHIFT ),
    m_phi( DEFAULT_PHI<<F_SHIFT ),
    m_eyedist( DEFAULT_EYEDIST )
{
  reCalc();
}

void CCamera::reCalc( void )
{
  // recalculate eye
}

CCamera::CCamera( CVector vCenter, fixed p, fixed theta, fixed phi, fixed gamma,
		  long eyedist)
  : m_vCenter( vCenter ),
    m_p( p ),
    m_theta( theta<<F_SHIFT ),
    m_gamma( gamma<<F_SHIFT ),
    m_phi( phi<<F_SHIFT ),
    m_eyedist( eyedist )
{
  reCalc();
}

CCamera::~CCamera( )
{
}

void CCamera::changeViewingAngle(fixed theta, fixed phi = 0, fixed gamma = 0)
{
  m_theta = theta<<F_SHIFT;
  m_phi = phi<<F_SHIFT;
  m_gamma = gamma<<F_SHIFT;

  reCalc();
}

void CCamera::shiftViewingAngle( fixed theta, fixed phi = 0, fixed gamma = 0 )
{
  m_theta = m_theta + (theta<<F_SHIFT);
  m_theta = m_theta % (NUM_ANGLES << F_SHIFT);

  m_phi = phi + (phi<<F_SHIFT);
  m_phi = m_phi % (NUM_ANGLES << F_SHIFT);

  m_gamma = gamma + (gamma<<F_SHIFT);
  m_gamma = m_gamma % (NUM_ANGLES << F_SHIFT);

  reCalc();
}

void CCamera::changeViewingDistance(fixed p)
{
  m_p = p;

  reCalc();
}

void CCamera::shiftViewingDistance(fixed dp)
{
  m_p += dp;

  reCalc();
}

void CCamera::changeFocus(const CVector &m_vNormalewCenter)
{
  m_vCenter = m_vNormalewCenter;

  reCalc();
}

void CCamera::changeEyeDistance(fixed newEyeDist)
{
  m_eyedist = newEyeDist;

  reCalc();
}

bool CCamera::convertWorldToScreen ( CVector vPt, long & ScreenX, long & ScreenY ) const
{
  ScreenX = long(SCR_X/2 + (m_eyedist*vPt.m_x / (vPt.m_z+m_p)));
  ScreenY = long(SCR_Y/2 - (m_eyedist*vPt.m_y / (vPt.m_z+m_p)));

  // assume it's always in front of us.  we'll worry about checking later
  return true;
}

CVector CCamera::convertWorldToCamWorld ( CVector vWorldPoint ) const
  // converts a world-coordinate point to camera-world-coordinate point
  // by taking into account camera location
{
  CVector vNew, vOld;
  fixed theta, phi, gamma;

  theta = -(m_theta>>F_SHIFT);
  if (theta < 256)
    theta += 256;
  theta &= 0xFF;

  phi = -(m_phi>>F_SHIFT);
  if (phi < 256)
    phi += 256;
  phi &= 0xFF;

  gamma = -(m_gamma>>F_SHIFT);
  if (gamma < 256)
    gamma += 256;
  gamma &= 0xFF;


  vOld = vWorldPoint - this->m_vCenter;     // first shift it

  // next, rotate it

  // rotate about x
  vNew.m_y = long((vOld.m_y*Global.m_cosLookup[theta]) - (vOld.m_z*Global.m_sinLookup[theta]))>>F_SHIFT;
  vNew.m_z = long((vOld.m_z*Global.m_cosLookup[theta]) + (vOld.m_y*Global.m_sinLookup[theta]))>>F_SHIFT;

  vOld.m_y = vNew.m_y;
  vOld.m_z = vNew.m_z;

  // rotate about y
  vNew.m_z = long((vOld.m_z*Global.m_cosLookup[phi]) - (vOld.m_x*Global.m_sinLookup[phi]))>>F_SHIFT;
  vNew.m_x = long((vOld.m_x*Global.m_cosLookup[phi]) + (vOld.m_z*Global.m_sinLookup[phi]))>>F_SHIFT;

  vOld.m_x = vNew.m_x;
  vOld.m_z = vNew.m_z;

  // rotate about z
  vNew.m_x = long((vOld.m_x*Global.m_cosLookup[gamma]) - (vOld.m_y*Global.m_sinLookup[gamma]))>>F_SHIFT;
  vNew.m_y = long((vOld.m_y*Global.m_cosLookup[gamma]) + (vOld.m_x*Global.m_sinLookup[gamma]))>>F_SHIFT;

  return vNew;
}

// show contents of Camera view on buffer
void CCamera::view( CVideo &rVideo, const CPolyObj & PolyObj )
{
  int i, j, k;
  CVector *vCamWorld;
  CVector vNew, vOld;

  // edge tracing section
  long ScreenX[3], ScreenY[3];
  CVector v1, v2;

  // for each polyobj's poly
  for (j = 0; j < PolyObj.m_NumPolys; j++)
    {
      // determine if we can even see this polygon
      CVector PointA = convertWorldToCamWorld( PolyObj.m_pPolys[j].m_pWorldPoints[ 0 ] );
      CVector PointB = convertWorldToCamWorld( PolyObj.m_pPolys[j].m_pWorldPoints[ 1 ] );
      CVector PointC = convertWorldToCamWorld( PolyObj.m_pPolys[j].m_pWorldPoints[ 2 ] );

      convertWorldToScreen( PointA, ScreenX[0], ScreenY[0] );
      convertWorldToScreen( PointB, ScreenX[1], ScreenY[1] );
      convertWorldToScreen( PointC, ScreenX[2], ScreenY[2] );
      
      v1.m_x = ScreenX[1] - ScreenX[0];
      v1.m_y = ScreenY[1] - ScreenY[0];
      v2.m_x = ScreenX[2] - ScreenX[1];
      v2.m_y = ScreenY[2] - ScreenY[1];
      
      /*
                             /\ 2                      \
                           /    \                        \ v2
                         /       \ 1                      \
                       /        /     the poly           /       the vectors
                        \     /                        / v1
                          \ / 0                      /
      */

      // get cross product (just z value, though)
      if ((v1.m_x*v2.m_y - v1.m_y*v2.m_x) > 0)
        continue; // skip it since facingaway from us



      // texture-mapping goes here...

      
      // store in edge array
      long StartX, StartY,  EndX, EndY;
      struct {
        long m_x, m_u,m_v;
      } l_edge[200], r_edge[200];
      
      for (k = 0; k < 200; k++)
	{
	  // init edges
	  l_edge[k].m_x = 320;
	  r_edge[k].m_x = -1;
	}

      CPoly ScreenPoly;
      convert3dPolyToScreenPoly( PolyObj.m_pPolys[j], ScreenPoly );

      // clip polygon here
      CPoly NewPoly;
      clipPolyScreen( ScreenPoly, NewPoly );

      /*      for (k = 0; k < NewPoly.m_NumPoints; k++)
	{
	  int & NumPoints = NewPoly.m_NumPoints;
	  printf("drawing line from %f,%f to %f,%f\n", 
		 NewPoly.m_pPoints[ k ].m_x, 
		 NewPoly.m_pPoints[ k ].m_y,  
		 NewPoly.m_pPoints[ (k + 1) % NumPoints ].m_x, 
		 NewPoly.m_pPoints[ (k + 1) % NumPoints ].m_y );

	  
	  rVideo.drawLine( NewPoly.m_pPoints[ k ].m_x, 
			   NewPoly.m_pPoints[ k ].m_y,  
			   NewPoly.m_pPoints[ (k + 1) % NumPoints ].m_x, 
			   NewPoly.m_pPoints[ (k + 1) % NumPoints ].m_y,  
			   15 );
	}
*/
      
      // store each point
      for (k = 0; k < NewPoly.m_NumPoints; k++)
	{
	  int NumPoints = NewPoly.m_NumPoints;

	  //CVector vFrom = convertWorldToCamWorld( NewPoly.m_pPoints[ k ] );
	  //CVector vTo = convertWorldToCamWorld( NewPoly.m_pPoints[ (k+1)%NumPoints ] );
	  //convertWorldToScreen( vFrom, StartX,StartY );
	  //convertWorldToScreen( vTo, EndX,EndY );

	  StartX = NewPoly.m_pPoints[ k ].m_x;
	  StartY = NewPoly.m_pPoints[ k ].m_y;
	  EndX = NewPoly.m_pPoints[ (k+1)%NumPoints ].m_x;
	  EndY = NewPoly.m_pPoints[ (k+1)%NumPoints ].m_y;

	  //rVideo.drawLine( StartX,StartY,  EndX,EndY, 14 );


	  // since our line endpoints are set up counterclockwise, if the second
	  // line's y is less than the first line's y, it must be a left edge
	  // otherwise, it's a right edge
	  long x, y, dx, dy,  u, v, du, dv, startu, startv, endu,endv, steps;
	  
	  steps = abs(EndY - StartY) + 1;
	  dx = ((EndX-StartX)<<F_SHIFT) / steps;
	  x = StartX<<F_SHIFT;
	  if (StartY < EndY)
	    dy = 1;
	  else
	    dy = -1;
	  
	  // texture coords
	  startu = NewPoly.m_pTextureMaps[ k ].m_u << F_SHIFT;
	  startv = NewPoly.m_pTextureMaps[ k ].m_v << F_SHIFT;
	  endu = NewPoly.m_pTextureMaps[ (k+1)%NumPoints ].m_u << F_SHIFT;
	  endv = NewPoly.m_pTextureMaps[ (k+1)%NumPoints ].m_v << F_SHIFT;

	  du = (endu - startu)/steps;
	  dv = (endv - startv)/steps;
	  
	  u = startu;
	  v = startv;
	  for (y = StartY; y != EndY; y += dy)
	    {
	      if (y >= 0 && y < 200)
		{
		  if (EndY > StartY)
		    {
		      if ( (x>>F_SHIFT) >= 0)
			l_edge[y].m_x = x>>F_SHIFT;
		      else
			l_edge[y].m_x = 0;
		      l_edge[y].m_u = u;
		      l_edge[y].m_v = v;
		    }
		  else
		    {
		      if ( (x>>F_SHIFT) < 320)
			r_edge[y].m_x = x>>F_SHIFT;
		      else
			r_edge[y].m_x = 319;
		      r_edge[y].m_u = u;
		      r_edge[y].m_v = v;
		    }
		}
	      x += dx;
	      u += du;
	      v += dv;
	    }
	}
      
      for (k = 0; k < 200; k++)
	{
	  long tmpx, tmpu, tmpv;
	  
	  if (r_edge[k].m_x != -1 && l_edge[k].m_x != 320 && r_edge[k].m_x < l_edge[k].m_x)
	    {
	      tmpx = l_edge[k].m_x;
	      tmpu = l_edge[k].m_u;
	      tmpv = l_edge[k].m_v;
	      
	      l_edge[k].m_x = r_edge[k].m_x;
	      l_edge[k].m_u = r_edge[k].m_u;
	      l_edge[k].m_v = r_edge[k].m_v;
	      
	      r_edge[k].m_x = tmpx;
	      r_edge[k].m_u = tmpu;
	      r_edge[k].m_v = tmpv;
	    }
	}
      
      // draw each line
      for (k = 0; k < 200; k++)
	{
	  // skip it if no edges there at all
	  if (l_edge[k].m_x == 320 || r_edge[k].m_x == -1)
	    continue;
	  
	  long u, du, v, dv;
	  
	  // draw texture there
	  u = l_edge[k].m_u;
	  du = (r_edge[k].m_u - l_edge[k].m_u)/(r_edge[k].m_x - l_edge[k].m_x + 1);
	  v = l_edge[k].m_v;
	  dv = (r_edge[k].m_v - l_edge[k].m_v)/(r_edge[k].m_x - l_edge[k].m_x + 1);
	  
	  for (int n = l_edge[k].m_x; n <= r_edge[k].m_x; n++)
	    {
	      //	      rVideo.m_pVidBuffer[320*k + n] = PolyObj.m_ppTexture[ m_TextureId ]->m_pTexture[ PolyObj.m_ppTexture[ NewPoly.m_TextureId ]->m_uDim*(v>>F_SHIFT) + (u>>F_SHIFT)];

	      rVideo.m_pVidBuffer[320*k + n] = PolyObj.m_ppTexture[ 0 ]->m_pTexture[ PolyObj.m_ppTexture[ 0 ]->m_uDim*(v>>F_SHIFT) + (u>>F_SHIFT)];

	      //          rVideo.m_pVidBuffer[320*k + n] = 15;
	      u += du;
	      v += dv;
	    }
	}
    }
}


void CCamera::convert3dPolyToScreenPoly( const CPoly & OldPoly, CPoly & NewPoly ) const
{
  int NumPoints = OldPoly.m_NumPoints;

  NewPoly.m_NumPoints = NumPoints;
  NewPoly.m_pPoints = new CVector[ NumPoints ];
  NewPoly.m_pTextureMaps = new texture_map[ NumPoints ];

  assert( OldPoly.m_pWorldPoints != 0 );

  for ( int k = 0; k < NumPoints; k++ )
    {
      long X, Y;

      CVector vPt = convertWorldToCamWorld( OldPoly.m_pWorldPoints[ k ] );
      convertWorldToScreen( vPt, X,Y );

      NewPoly.m_pPoints[ k ].setPoint( X, Y, 0 ); 
      NewPoly.m_pTextureMaps[ k ] = OldPoly.m_pTextureMaps[ k ];
    }
}

long getHorizIntersection( double Y,  double x1, double y1,  double x2, double y2 )
{
  if ( y1 == y2 )
    return ((x2 + x1)/2);
  else
    {
      long x = ( ( ( Y - y1 ) * ( x2 - x1 )) / ( y2 - y1 ) + x1 );

      //assert( x < 320 );
      
      return x;
    }
}


void getUV( const CPoly & Poly, int Point1, int Point2, double x, double y, int & u, int & v )
{
  double x1,y1,  x2,y2;
  int u1,v1,  u2,v2;
  x1 = Poly.m_pPoints[ Point1 ].m_x;
  y1 = Poly.m_pPoints[ Point1 ].m_y;
  x2 = Poly.m_pPoints[ Point2 ].m_x;
  y2 = Poly.m_pPoints[ Point2 ].m_y;
  u1 = Poly.m_pTextureMaps[ Point1 ].m_u;
  v1 = Poly.m_pTextureMaps[ Point1 ].m_v;
  u2 = Poly.m_pTextureMaps[ Point2 ].m_u;
  v2 = Poly.m_pTextureMaps[ Point2 ].m_v;

  double FullDistance, PartDistance;

  FullDistance = sqrt(  SQR(x2-x1) + SQR(y2-y1) );
  PartDistance = sqrt(  SQR(x-x1) +  SQR(y-y1) );

  int Percent = (PartDistance*100)/FullDistance;

  u = ((u2-u1) * Percent)/100 + u1;
  v = ((v2-v1) * Percent)/100 + v1;


  /*

  if ( x2 != x1 )
    u = (( x - x1 ) * ( u2 - u1 )) / ( x2 - x1 ) + u1;
  else
    {
      printf("x2 == x1. x1,x2 = %f  x = %f  u1 = %d.  u2 = %d\n", x1, x, u1, u2);
      u = (u1 + u2)/2;
    }

  if ( y2 != y1 )
    v =( ( y - y1 ) * ( v2 - v1 )) / ( y2 - y1 ) + v1;
  else
    {
      printf("y2 == y1.  v1 = %d.  v2 = %d\n", v1, v2);
      v = (v1 + v2)/2;
    }

  */
}


void CCamera::clipPoly( const CPoly & OldPoly, CPoly & NewPoly, eClipType ClipType ) const 
{
  NewPoly.m_pPoints = new CVector[ OldPoly.m_NumPoints + 1 ];
  NewPoly.m_pTextureMaps = new texture_map[ OldPoly.m_NumPoints + 1 ];
  NewPoly.m_TextureId = OldPoly.m_TextureId;

  bool *IsInside = new bool[ OldPoly.m_NumPoints ];

  for ( int k = 0; k < OldPoly.m_NumPoints; k++ )
    {
      // calculate to see if this point is inside or outside the vector vC

      switch ( ClipType )
	{
	case CT_LEFT :
	  if ( OldPoly.m_pPoints[ k ].m_x <= LEFT_X )
	    IsInside[ k ] = false;
	  else
	    IsInside[ k ] = true;
	  break;
	case CT_RIGHT :
	  if ( OldPoly.m_pPoints[ k ].m_x >= RIGHT_X )
	    IsInside[ k ] = false;
	  else
	    IsInside[ k ] = true;
	  break;
	case CT_TOP :
	  if ( OldPoly.m_pPoints[ k ].m_y <= TOP_Y )
	    IsInside[ k ] = false;
	  else
	    IsInside[ k ] = true;
	  break;
	case CT_BOTTOM :
	  if ( OldPoly.m_pPoints[ k ].m_y >= BOTTOM_Y )
	    IsInside[ k ] = false;
	  else
	    IsInside[ k ] = true;
	  break;
	}			
    }

  int NumPoints = 0;

  int k;
  for ( k = 1; k < OldPoly.m_NumPoints; k++ )
    {
      // if this point is inside and previous was inside
      if ( IsInside[ k ] && IsInside[ k-1 ] )
	{
	  //printf("Going from inside to inside\n");

	  NewPoly.m_pPoints[ NumPoints ] = OldPoly.m_pPoints[ k ];
	  NewPoly.m_pTextureMaps[ NumPoints ] = OldPoly.m_pTextureMaps[ k ];
	  NumPoints++;
	}

      // if this point is outside and the previous was inside
      else if ( !IsInside[ k ] && IsInside[ k-1 ] )
	{
	  //printf("Going from inside to outside\n");

	  long x, y;
	  int u, v;
	  CVector NewPoint;

	  switch ( ClipType )
	    {
	    case CT_TOP :
	      x = getHorizIntersection( TOP_Y,
					OldPoly.m_pPoints[ k-1 ].m_x,
					OldPoly.m_pPoints[ k-1 ].m_y,
					OldPoly.m_pPoints[ k ].m_x,
					OldPoly.m_pPoints[ k ].m_y );
	      y = TOP_Y;
	      
	      NewPoint.setPoint( x, y, 0 );
	      break;
	    case CT_BOTTOM :
	      x = getHorizIntersection( BOTTOM_Y,
					OldPoly.m_pPoints[ k-1 ].m_x,
					OldPoly.m_pPoints[ k-1 ].m_y,
					OldPoly.m_pPoints[ k ].m_x,
					OldPoly.m_pPoints[ k ].m_y );
	      y = BOTTOM_Y;
	      
	      NewPoint.setPoint( x, y, 0 );
	      break;
	    case CT_LEFT :
	      y = getHorizIntersection( LEFT_X,
					OldPoly.m_pPoints[ k-1 ].m_y,
					OldPoly.m_pPoints[ k-1 ].m_x,
					OldPoly.m_pPoints[ k ].m_y,
					OldPoly.m_pPoints[ k ].m_x );
	      x = LEFT_X;
	      
	      NewPoint.setPoint( x, y, 0 );
	      break;
	    case CT_RIGHT :
	      y = getHorizIntersection( RIGHT_X,
					OldPoly.m_pPoints[ k-1 ].m_y,
					OldPoly.m_pPoints[ k-1 ].m_x,
					OldPoly.m_pPoints[ k ].m_y,
					OldPoly.m_pPoints[ k ].m_x );
	      x = RIGHT_X;
	      
	      NewPoint.setPoint( x, y, 0 );
	      break;
	    }
	  NewPoly.m_pPoints[ NumPoints ] = NewPoint;

	  //   interpolate texture for intersection as well...
	  getUV( OldPoly, k-1, k, x, y, u, v );
	  NewPoly.m_pTextureMaps[ NumPoints ].m_u = u;
	  NewPoly.m_pTextureMaps[ NumPoints ].m_v = v;

	  NumPoints++;
	}

      // if this point is outside and previous outside
      else if ( !IsInside[ k ] && !IsInside[ k-1 ] )
	{
	  //printf("Going from outside to outside\n");
	  //   do nothing
	}

      // if this point is inside and previous was outside
      else if ( IsInside[ k ] && !IsInside[ k-1 ] )
	{
	  //printf("Going from outside to inside\n");

	  long x, y;
	  int u, v;
	  CVector NewPoint;

	  switch ( ClipType )
	    {
	    case CT_TOP :
	      x = getHorizIntersection( TOP_Y,
					OldPoly.m_pPoints[ k-1 ].m_x,
					OldPoly.m_pPoints[ k-1 ].m_y,
					OldPoly.m_pPoints[ k ].m_x,
					OldPoly.m_pPoints[ k ].m_y );
	      y = TOP_Y;
	      
	      NewPoint.setPoint( x, y, 0 );
	      break;
	    case CT_BOTTOM :
	      x = getHorizIntersection( BOTTOM_Y,
					OldPoly.m_pPoints[ k-1 ].m_x,
					OldPoly.m_pPoints[ k-1 ].m_y,
					OldPoly.m_pPoints[ k ].m_x,
					OldPoly.m_pPoints[ k ].m_y );
	      y = BOTTOM_Y;
	      
	      NewPoint.setPoint( x, y, 0 );
	      break;
	    case CT_LEFT :
	      y = getHorizIntersection( LEFT_X,
					OldPoly.m_pPoints[ k-1 ].m_y,
					OldPoly.m_pPoints[ k-1 ].m_x,
					OldPoly.m_pPoints[ k ].m_y,
					OldPoly.m_pPoints[ k ].m_x );
	      x = LEFT_X;
	      
	      NewPoint.setPoint( x, y, 0 );
	      break;
	    case CT_RIGHT :
	      y = getHorizIntersection( RIGHT_X,
					OldPoly.m_pPoints[ k-1 ].m_y,
					OldPoly.m_pPoints[ k-1 ].m_x,
					OldPoly.m_pPoints[ k ].m_y,
					OldPoly.m_pPoints[ k ].m_x );
	      x = RIGHT_X;
	      
	      NewPoint.setPoint( x, y, 0 );
	      break;
	    }
	  NewPoly.m_pPoints[ NumPoints ] = NewPoint;

	  //   interpolate texture for intersection as well...
	  getUV( OldPoly, k-1, k, x, y, u, v );
	  NewPoly.m_pTextureMaps[ NumPoints ].m_u = u;
	  NewPoly.m_pTextureMaps[ NumPoints ].m_v = v;

	  NumPoints++;

	  //   add this point
	  NewPoly.m_pPoints[ NumPoints ] = OldPoly.m_pPoints[ k ];
	  NewPoly.m_pTextureMaps[ NumPoints ] = OldPoly.m_pTextureMaps[ k ];
	  NumPoints++;
	}
    }




      // if this point is inside and previous was inside
      if ( IsInside[ 0 ] && IsInside[ k-1 ] )
	{
	  //printf("Going from inside to inside\n");

	  NewPoly.m_pPoints[ NumPoints ] = OldPoly.m_pPoints[ 0 ];
	  NewPoly.m_pTextureMaps[ NumPoints ] = OldPoly.m_pTextureMaps[ 0 ];
	  NumPoints++;
	}

      // if this point is outside and the previous was inside
      else if ( !IsInside[ 0 ] && IsInside[ k-1 ] )
	{
	  //printf("Going from inside to outside\n");

	  long x, y;
	  int u, v;
	  CVector NewPoint;

	  switch ( ClipType )
	    {
	    case CT_TOP :
	      x = getHorizIntersection( TOP_Y,
					OldPoly.m_pPoints[ k-1 ].m_x,
					OldPoly.m_pPoints[ k-1 ].m_y,
					OldPoly.m_pPoints[ 0 ].m_x,
					OldPoly.m_pPoints[ 0 ].m_y );
	      y = TOP_Y;
	      
	      NewPoint.setPoint( x, y, 0 );
	      break;
	    case CT_BOTTOM :
	      x = getHorizIntersection( BOTTOM_Y,
					OldPoly.m_pPoints[ k-1 ].m_x,
					OldPoly.m_pPoints[ k-1 ].m_y,
					OldPoly.m_pPoints[ 0 ].m_x,
					OldPoly.m_pPoints[ 0 ].m_y );
	      y = BOTTOM_Y;
	      
	      NewPoint.setPoint( x, y, 0 );
	      break;
	    case CT_LEFT :
	      y = getHorizIntersection( LEFT_X,
					OldPoly.m_pPoints[ k-1 ].m_y,
					OldPoly.m_pPoints[ k-1 ].m_x,
					OldPoly.m_pPoints[ 0 ].m_y,
					OldPoly.m_pPoints[ 0 ].m_x );
	      x = LEFT_X;
	      
	      NewPoint.setPoint( x, y, 0 );
	      break;
	    case CT_RIGHT :
	      y = getHorizIntersection( RIGHT_X,
					OldPoly.m_pPoints[ k-1 ].m_y,
					OldPoly.m_pPoints[ k-1 ].m_x,
					OldPoly.m_pPoints[ 0 ].m_y,
					OldPoly.m_pPoints[ 0 ].m_x );
	      x = RIGHT_X;
	      
	      NewPoint.setPoint( x, y, 0 );
	      break;
	    }
	  NewPoly.m_pPoints[ NumPoints ] = NewPoint;

	  //   interpolate texture for intersection as well...
	  getUV( OldPoly, k-1, 0, x, y, u, v );
	  NewPoly.m_pTextureMaps[ NumPoints ].m_u = u;
	  NewPoly.m_pTextureMaps[ NumPoints ].m_v = v;
	  NumPoints++;
	}

      // if this point is outside and previous outside
      else if ( !IsInside[ 0 ] && !IsInside[ k-1 ] )
	{
	  //printf("Going from outside to outside\n");
	  //   do nothing
	}

      // if this point is inside and previous was outside
      else if ( IsInside[ 0 ] && !IsInside[ k-1 ] )
	{
	  ////printf("Going from outside to inside\n");

	  long x, y;
	  int u, v;
	  CVector NewPoint;

	  switch ( ClipType )
	    {
	    case CT_TOP :
	      x = getHorizIntersection( TOP_Y,
					OldPoly.m_pPoints[ k-1 ].m_x,
					OldPoly.m_pPoints[ k-1 ].m_y,
					OldPoly.m_pPoints[ 0 ].m_x,
					OldPoly.m_pPoints[ 0 ].m_y );
	      y = TOP_Y;
	      
	      NewPoint.setPoint( x, y, 0 );
	      break;
	    case CT_BOTTOM :
	      x = getHorizIntersection( BOTTOM_Y,
					OldPoly.m_pPoints[ k-1 ].m_x,
					OldPoly.m_pPoints[ k-1 ].m_y,
					OldPoly.m_pPoints[ 0 ].m_x,
					OldPoly.m_pPoints[ 0 ].m_y );
	      y = BOTTOM_Y;
	      
	      NewPoint.setPoint( x, y, 0 );
	      break;
	    case CT_LEFT :
	      y = getHorizIntersection( LEFT_X,
					OldPoly.m_pPoints[ k-1 ].m_y,
					OldPoly.m_pPoints[ k-1 ].m_x,
					OldPoly.m_pPoints[ 0 ].m_y,
					OldPoly.m_pPoints[ 0 ].m_x );
	      x = LEFT_X;
	      
	      NewPoint.setPoint( x, y, 0 );
	      break;
	    case CT_RIGHT :
	      y = getHorizIntersection( RIGHT_X,
					OldPoly.m_pPoints[ k-1 ].m_y,
					OldPoly.m_pPoints[ k-1 ].m_x,
					OldPoly.m_pPoints[ 0 ].m_y,
					OldPoly.m_pPoints[ 0 ].m_x );
	      x = RIGHT_X;
	      
	      NewPoint.setPoint( x, y, 0 );
	      break;
	    }
	  NewPoly.m_pPoints[ NumPoints ] = NewPoint;

	  //   interpolate texture for intersection as well...
	  getUV( OldPoly, k-1, 0, x, y, u, v );
	  NewPoly.m_pTextureMaps[ NumPoints ].m_u = u;
	  NewPoly.m_pTextureMaps[ NumPoints ].m_v = v;
	  NumPoints++;


	  //   add this point
	  NewPoly.m_pPoints[ NumPoints ] = OldPoly.m_pPoints[ 0 ];
	  NewPoly.m_pTextureMaps[ NumPoints ] = OldPoly.m_pTextureMaps[ 0 ];
	  NumPoints++;
	}


  NewPoly.m_NumPoints = NumPoints;
}


void CCamera::clipPolyScreen( const CPoly & OldPoly, CPoly & NewPoly ) const 
{
  CPoly PolyA, PolyB, PolyC;

  clipPoly( OldPoly, PolyA, CT_LEFT );
  clipPoly( PolyA, PolyB, CT_RIGHT );
  clipPoly( PolyB, PolyC, CT_BOTTOM );
  clipPoly( PolyC, NewPoly, CT_TOP );

  //printf("Num points: %d\n", NewPoly.m_NumPoints);
}
